@@ -82,7 +82,6 @@ class LayersGradTest(unittest.TestCase):
             vanilla_run_grads, private_run_grads
         ):
             assert vanilla_name == private_name
-
             self.assertTrue(
                 torch.allclose(vanilla_grad, private_grad, atol=10e-5, rtol=10e-3),
                 f"Gradient mismatch. Parameter: {layer}.{vanilla_name}, loss: {criterion.reduction}",
@@ -162,8 +161,10 @@ class LayersGradTest(unittest.TestCase):
 
     def test_embedding(self):
         layer = nn.Embedding(256, 100)
-        x1 = torch.randint(0, 255, (128, 42)).long()
+        x0 = torch.randint(0, 255, (16, 8, 4)).long()
+        x1 = torch.randint(0, 255, (16, 8)).long()
         x2 = torch.randint(0, 255, (64,)).long()
+        self._check_one_layer(layer, x0)
         self._check_one_layer(layer, x1)
         self._check_one_layer(layer, x2)
 
